#include "LightSourceProperties.h"
#include "GLErrorHandler.h"
#include <cstdio>  // für printf

GLenum LightSourceProperties::LIGHT_ID[] = { GL_LIGHT0, GL_LIGHT1, GL_LIGHT2, GL_LIGHT3,
                                        GL_LIGHT4, GL_LIGHT5, GL_LIGHT6, GL_LIGHT7 };
                                        
LightSourceProperties::LightSourceProperties () :
      lightType(DIRECTIONAL),
      isOn(false),
      isFixed(false),
      ambientColor(ColorSpec(0.0f, 0.0f, 0.0f, 1.0f)),
      diffuseColor(ColorSpec(0.0f, 0.0f, 0.0f, 1.0f)),
      specularColor(ColorSpec(0.0f, 0.0f, 0.0f, 1.0f)),
      direction(SpacialPoint(0.0f, 0.0f, 1.0f)),
      constantAttenuation(1.0f),
      linearAttenuation(0.0f),
      quadraticAttenuation(0.0f),
      spotDirection(SpacialPoint(0.0f, 0.0f, 11.0f)),
      exponent(0.0f),
      cutoff(180.0f)

{
}

void LightSourceProperties::setLightID(GLenum id) {
    lightID = id;
}

void LightSourceProperties::isFixedLight(bool fl) {
	isFixed = fl;
}

bool LightSourceProperties::isFixedLight() {
	return isFixed;
}

ColorSpec* LightSourceProperties::getPtrToAmbientColor() {
	return &ambientColor;
}

ColorSpec* LightSourceProperties::getPtrToDiffuseColor() {
	return &diffuseColor;
}

ColorSpec* LightSourceProperties::getPtrToSpecularColor() {
	return &specularColor;
}

void LightSourceProperties::installLight() {
    glLightf(lightID, GL_SPOT_EXPONENT, 0.0f);
    glLightf(lightID, GL_SPOT_CUTOFF, 180.0f);
    glLightf(lightID, GL_CONSTANT_ATTENUATION, 1.0f);
    glLightf(lightID, GL_LINEAR_ATTENUATION, 0.0f);
    glLightf(lightID, GL_QUADRATIC_ATTENUATION, 0.0f);
    
    glLightfv(lightID, GL_AMBIENT, ambientColor.getColorSettings());
    glLightfv(lightID, GL_DIFFUSE, diffuseColor.getColorSettings());
    glLightfv(lightID, GL_SPECULAR, specularColor.getColorSettings());
    
    GLfloat vec4[4];  //  müssen u.U. Instanzvariablen sein, siehe Erfahrung
                      //  mit Projektionsmatrizen
    GLfloat vec3[3];
    
    vec4[0] = direction.x;
    vec4[1] = direction.y;
    vec4[2] = direction.z;
    vec4[3] = lightType == DIRECTIONAL ? 0.0f : 1.0f;        //  directional light
    glLightfv(lightID, GL_POSITION, vec4);
    
    if (lightType == DIRECTIONAL) {
		return;
	}
	//printf("Dämpfung %d   %f   %f  %f\n", lightID, constantAttenuation, linearAttenuation, quadraticAttenuation);
    glLightf(lightID, GL_CONSTANT_ATTENUATION, constantAttenuation);
    GLErrorHandler::checkError("LightSourceProperties", "installTransformedLight", "constantAttenuation");
    glLightf(lightID, GL_LINEAR_ATTENUATION, (GLfloat) linearAttenuation);
    GLErrorHandler::checkError("LightSourceProperties", "installLight", "linearAttenuation");
    glLightf(lightID, GL_QUADRATIC_ATTENUATION, quadraticAttenuation);
    GLErrorHandler::checkError("LightSourceProperties", "installTransformedLight", "quadrativAttenuation");
    
    if (lightType == POSITIONAL) {
		return;
	}

    vec3[0] = spotDirection.x;
    vec3[1] = spotDirection.y;
    vec3[2] = spotDirection.z;
    glLightfv(lightID, GL_SPOT_DIRECTION, vec3);
    glLightf(lightID, GL_SPOT_EXPONENT, exponent);
    glLightf(lightID, GL_SPOT_CUTOFF, cutoff);
}


/*
 * Hier ist zu prüfen, ob die Transformation der Orbjektkoordinaten in
 * Kamerakoordinaten nicht einfacher bewerkstelligt werden kann. 
 * Es sind sowohl Positionen als auch Richtungen zu transformieren, und
 * zwar in unterschiedlicher Art!
 */
void LightSourceProperties::installTransformedLight(glm::mat4 &modelViewMatrix){
	
    glLightf(lightID, GL_SPOT_EXPONENT, 0.0f);
    glLightf(lightID, GL_SPOT_CUTOFF, 180.0f);
    
	glLightfv(lightID, GL_AMBIENT, ambientColor.getColorSettings());
    glLightfv(lightID, GL_DIFFUSE, diffuseColor.getColorSettings());
    glLightfv(lightID, GL_SPECULAR, specularColor.getColorSettings());
    
    GLfloat vec4[4];  //  müssenu.U. Instanzvariablen sein, siehe Erfahrung
                    //  mit Projektionsmatrizen
    GLfloat vec3[3];
    
    if (lightType == DIRECTIONAL) {
		vec4[0] = direction.x;
		vec4[1] = direction.y;
		vec4[2] = direction.z;
		vec4[3] =  0.0f;        //  directional light
	} else {
		glm::vec4 p, tp;
		p[0] = direction.x;
		p[1] = direction.y;
		p[2] = direction.z;
		p[3] = 1.0f; 
		tp = modelViewMatrix * p; 
		//printf("Objektkoordinaten  %f   %f   %f   %f\n", p[0], p[1], p[2], p[3]);
		//printf("Kamerakoordinaten  %f   %f   %f   %f\n", tp[0], tp[1], tp[2], tp[3]);
		//printf("------------------------------\n");
		vec4[0] = tp[0];
		vec4[1] = tp[1];
		vec4[2] = tp[2];
		vec4[3] = 1.0f;        //  positional light
	}
    glLightfv(lightID, GL_POSITION, vec4);
    
    if (lightType == DIRECTIONAL) {
		return;
	}
	
    glLightf(lightID, GL_CONSTANT_ATTENUATION, constantAttenuation);
    GLErrorHandler::checkError("LightSourceProperties", "installTransformedLight", "constantAttenuation");
    glLightf(lightID, GL_LINEAR_ATTENUATION, linearAttenuation);
    GLErrorHandler::checkError("LightSourceProperties", "installTransformedLight", "linearAttenuation");
    glLightf(lightID, GL_QUADRATIC_ATTENUATION, quadraticAttenuation);
    GLErrorHandler::checkError("LightSourceProperties", "installTransformedLight", "quadraticAttenuation");
    
    if (lightType == POSITIONAL) {
		return;
	}
	
	
    vec3[0] = spotDirection.x;
    vec3[1] = spotDirection.y;
    vec3[2] = spotDirection.z;
    glLightfv(lightID, GL_SPOT_DIRECTION, vec3);
    glLightf(lightID, GL_SPOT_EXPONENT, exponent);
    glLightf(lightID, GL_SPOT_CUTOFF, cutoff);
}

void LightSourceProperties::setConstantAttenuation(float v) {
	constantAttenuation = v;
}

void LightSourceProperties::setLinearAttenuation(float v) {
	linearAttenuation = v;
}

void LightSourceProperties::setQuadraticAttenuation(float v) {
	quadraticAttenuation = v;
}
