#include "OpenGLSceneView.h"
#include "SpacialPoint.h"
#include "EGLInterface.h"
#include <cairomm/context.h>
#include <gdkmm/general.h>   //  set_source_pixbuf()
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtx/transform2.hpp>
#include <iostream>


using namespace std;

OpenGLSceneView::OpenGLSceneView(SceneHolder& sm) :
   sceneModel(sm)
{
    sceneModel.addAttributeObserver(this);
    state = &_state;
       // Clear application state
    memset( state, 0, sizeof( *state ) );
    this->set_double_buffered(false);   //  für native window
    EGLInterface::initEGL();
    EGLInterface::createEGLContext(state, (GtkWidget*) this -> gobj());
}

//  Alle EGL-Resourcen freigeben
OpenGLSceneView:: ~OpenGLSceneView() {
    sceneModel.removeAttributeObserver(this);
    EGLInterface::freeEGLContext(state);
    EGLInterface::shutdownEGL(state);
}



void OpenGLSceneView::init_model_proj(EGL_STATE_T* state) {

   glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

   glViewport(0, 0, (GLsizei)state->image_width, (GLsizei)state->image_width);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();

   reset_model(state);
}

void::OpenGLSceneView::reset_model(EGL_STATE_T* state) {
   // reset model position
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
   glEnable(GL_DEPTH_TEST);
   glDepthFunc(GL_LEQUAL);
   glEnable(GL_LINE_SMOOTH);
   glShadeModel(GL_SMOOTH);

}


void::OpenGLSceneView::redraw_scene(EGL_STATE_T* state){
    Scene* scene = sceneModel.getScene();
    if (scene != (Scene*) 0) {
		glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );  //  new
		glHint( GL_LINE_SMOOTH_HINT, GL_NICEST );  //  new
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        //glm::mat4 projection = glm::ortho();
        //  lt. Dokumentation ist hier die Projektion anzugeben
        scene -> installProjection();
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        glm::mat4 view =
            glm::lookAt(
                scene -> getCameraSettings() -> position.asVec3(),     //  Position der Kamera
                scene -> getCameraSettings() -> target.asVec3(),       //  target
                scene -> getCameraSettings() -> upDirection.asVec3()); //  up-vector

        scene -> installStage(view);  //  hier wird die Scenebox gesetzt
                                      //  ausserdem werden die beweglichen Lichter gesetzt

        glMultMatrixf(glm::value_ptr(view));
        // hier gehört die Rotationsmatrix her
        //glRotatef((float) state -> rot_angle_x, 1.0f, 0.0f, 0.0f);
        //glRotatef((float) state -> rot_angle_y, 0.0f, 1.0f, 0.0f);
        //glRotatef((float) state -> rot_angle_z, 0.0f, 0.0f, 1.0f);
        glMultMatrixf(glm::value_ptr(*(scene -> getCurrentTransformationMatrix())));
        scene -> renderScene();  // ortsfeste Lichter und Objekte darstellen
    }
    glFinish();
}



//
//  gesendet von SceneHolder
//
void OpenGLSceneView::onSceneAttributesChanged() {
    redraw_scene(state);
    queue_draw();
}

//
//  Diese Methode wird aufgerufen, wenn sich die Platzierung
//  oder die Groesse des Fensters aendert.
//  Vorhandene Puffer werden zurueckgegeben, neue Puffer in der
//  erforderlichen Groesse werden hergestellt.
//
bool OpenGLSceneView::on_configure_event(GdkEventConfigure* event) {
    int useableWidth = (get_width()/16) * 16;
    int useableHeight = (get_height()/16)*16;
    int s = useableWidth < useableHeight ? useableWidth : useableHeight;
    state -> screen_width = get_width();
    state -> screen_height = get_height();
    state -> image_width = s;

    EGLInterface::resizeEGLBuffer(state, (GtkWidget*) this -> gobj());

    // Setup the model world
    init_model_proj(state);
    redraw_scene(state);
    return true;
}

//
//  Hier wird das gespeicherte Bild lediglich in einen Pixbuf
//  kopiert und sodann angezeigt.
//
bool OpenGLSceneView::on_draw(const Cairo::RefPtr<Cairo::Context> &context) {
   eglSwapBuffers( state->display, state->surface);
   return true;
}


glm::mat4 OpenGLSceneView::rotationMatrix (float angle, SpacialPoint axis, SpacialPoint center) {
	glm::mat4 *tm = sceneModel.getScene() -> getCurrentTransformationMatrix();
    glm::mat4 tr1 =
            glm::translate(*tm, glm::vec3(center.x, center.y, center.z));
      // für neuere Versionen von glm ist im nachfolgenden Ausdruck
      //  statt   -angle  zu schreiben:    glm::radians(-angle)
    glm::mat4 rot(glm::rotate(tr1, glm::radians(-angle),
                         glm::vec3(axis.x, axis.y, axis.z)));
    glm::mat4 tr2(glm::translate(rot, glm::vec3(-center.x, -center.y, -center.z)));
    return tr2;
}

void OpenGLSceneView:: refresh() {
    state -> rot_angle_x = 0.0;
    state -> rot_angle_y = 0.0;
    state -> rot_angle_z = 0.0;
	redraw_scene(state);
	queue_draw();
}

void OpenGLSceneView::rotateX(double x) {
	double rX = x - state -> rot_angle_x;
    state -> rot_angle_x = x;
    Scene* scene = sceneModel.getScene();
    if (scene  != (Scene*) 0) {
        CameraSettings* cs = scene -> getCameraSettings();
        glm::mat4 transformation = rotationMatrix((float) rX,
                                                  SpacialPoint(1.0f, 0.0f, 0.0f),
                                                  cs -> target);

        glm::mat4 *tm = scene -> getCurrentTransformationMatrix();
        *tm = transformation;

        redraw_scene(state);
        queue_draw();
	}
}

void OpenGLSceneView::rotateY(double y) {
	double rY = y - state -> rot_angle_y;
    state -> rot_angle_y = y;
    Scene* scene = sceneModel.getScene();
    if (scene  != (Scene*) 0) {
        CameraSettings* cs = scene -> getCameraSettings();
        glm::mat4 transformation = rotationMatrix((float) rY,
                                                  SpacialPoint(0.0f, 1.0f, 0.0f),
                                                  cs -> target);

        glm::mat4 *tm = scene -> getCurrentTransformationMatrix();
        *tm = transformation;

        redraw_scene(state);
        queue_draw();
	}
}

void OpenGLSceneView::rotateZ(double z) {
	double rZ = z - state -> rot_angle_z;
    state -> rot_angle_z = z;
    Scene* scene = sceneModel.getScene();
    if (scene  != (Scene*) 0) {
        CameraSettings* cs = scene -> getCameraSettings();
        glm::mat4 transformation = rotationMatrix((float) rZ,
                                                  SpacialPoint(0.0f, 0.0f, 1.0f),
                                                  cs -> target);

        glm::mat4 *tm = scene -> getCurrentTransformationMatrix();
        *tm = transformation;

        redraw_scene(state);
        queue_draw();
	}
}

