#include "Scene.h"

#include "OrthographicProjectionSpec.h"
#include "PerspectiveProjectionSpec.h"

Scene::Scene(ColorSpec &clearC, ColorSpec &ambientL)
  :  clearColor(clearC),
     ambientLightColor(ambientL)
{
    for (int i = 0; i < NUMBER_OF_LIGHTS; i++) {
        glDisable(LightSourceProperties::LIGHT_ID[i]);
    }
}

Scene::~Scene() {
    while (!sceneObjects.empty()) {
        delete sceneObjects.front();
        sceneObjects.pop_front();
    }
    sceneObjects.clear();
    for (int i = 0; i < NUMBER_OF_LIGHTS; i++) {
        glDisable(LightSourceProperties::LIGHT_ID[i]);
    }
}

ColorSpec Scene::getAmbientLightColor() {
    return ambientLightColor;
}

ColorSpec& Scene::getAmbientLightColorRef() {
    return ambientLightColor;
}

CameraSettings* Scene::getCameraSettings() {
	return &cameraSettings;
}

LightSourceProperties* Scene::getRefToLight(int idx) {
	return &(lightSources[idx]);
}

void Scene::addProjectionSpec(AbstractProjectionSpec* p) {
	//delete projection;
	projection = p;
}

void Scene::addLightSources(LightSourceProperties *ls) {
	lightSources = ls;
}

ColorSpec Scene::getClearColor() {
    return clearColor;
}

ColorSpec& Scene::getClearColorRef() {
    return clearColor;
}

Sphere Scene::enclosingSphere() {
    return enclosingBox().enclosingSphere();
}

BoundingBox Scene::enclosingBox() {
    BoundingBox bb;
    for (std::list<AbstractSceneObject*>::iterator Iterator = sceneObjects.begin();
        Iterator != sceneObjects.end();
        ++Iterator) {
        (*Iterator) -> addEnclosingBox(&bb);
    }
        
    return bb;
}

SpacialPoint Scene::getCenterPoint() {
    return enclosingSphere().getCenterPoint();
}

void Scene::addSceneObject(AbstractSceneObject* obj) {
    sceneObjects.push_back(obj);
}

void Scene::installProjection() {
	projection -> setSpace(this);
}

void Scene::setTransformationMatrix(glm::mat4 *tm) {
	transformationMatrix = tm;
}

glm::mat4*Scene::getCurrentTransformationMatrix() {
	return transformationMatrix;
}

/*
 * Diese Methode wird vor der Festlegung der Transformationsmatrix
 * aufgerufen. Bewegliche Lichter werden hier gesetzt. Die Koordinaten
 * der beweglichen Lichter sind in eye-coordinates umzurechnen 
 */
void Scene::installStage(glm::mat4 &modelViewMatrix) {
    
    glClearDepthf(1.0f);
    glClearColor(clearColor.getRedf(), clearColor.getGreenf(),
                 clearColor.getBluef(), clearColor.getAlphaf());
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glEnable(GL_LIGHTING);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientLightColor.getColorSettings());
    glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);  //  GL_FALSE liefert andere Lichteffekte
     // added 2015-07-03  ortsfeste Lichter
     
     
    for (int i = 0; i < NUMBER_OF_LIGHTS; i++) {
        glDisable(LightSourceProperties::LIGHT_ID[i]);
        if (lightSources[i].isOn) {
            if (!lightSources[i].isFixedLight()) {
				lightSources[i].installTransformedLight(modelViewMatrix);
                //glEnable(LightSourceProperties::LIGHT_ID[i]);
			}
        } 
    }  
}


/*
 * Diese Methode wird nach der Festlegung der Transformationslichter
 * aufgerufen. Ortsfeste Lichter werden hier gesetzt
 */
void Scene::renderScene() {
    //  in der Szene platzierte, bewegte Lichter installieren

	for (int i= 0; i < NUMBER_OF_LIGHTS; i++) {
		if (lightSources[i].isOn) {
			if (lightSources[i].isFixedLight()) {
				lightSources[i].installLight();
				//glEnable(LightSourceProperties::LIGHT_ID[i]);
			}
		}
	}
    
    // Szeneobjekte zeichnen

    for (std::list<AbstractSceneObject*>::iterator Iterator = sceneObjects.begin();
         Iterator != sceneObjects.end();
         ++Iterator) {
        // Sichtbarkeit pruefen, Lichter setzen
        if ((*Iterator) -> isVisible()) {
			bool* lightFlags = (*Iterator) -> getLightSettings();
			for (int i = 0; i < NUMBER_OF_LIGHTS; i++) {
				if (lightSources[i].isOn) {
					if (lightFlags[i]) {
                        glEnable(LightSourceProperties::LIGHT_ID[i]);
					} else {
                        glDisable(LightSourceProperties::LIGHT_ID[i]);
					}
				}
			}
            glLightModelf(GL_LIGHT_MODEL_TWO_SIDE,
                          (*Iterator) -> useLightModelTwoSide()
                            ? GL_TRUE
                            : GL_FALSE); 
            (*Iterator) -> renderObject();
            //glFinish();   // 2017-05-05 Versuchsweise eingefügt, um
                            // selten auftretenden Fehler im Beispiel
                            // mit der sphärischenSchraubenlinie  zu
                            // vermeiden.
                            // 2017-05-23: Bei On-Screen-Rendering
                            // verursacht dies Flackern und Anzeige
                            // unfertiger Szenen.
		}
    }
}

std::list<AbstractSceneObject*>& Scene::getObjectList() {
	return sceneObjects;
}

